import numpy as np
from dataclasses import dataclass, field
import plotly.graph_objects as go
from plotly.subplots import make_subplots


light_temperature_map = [
    [0.0, "rgb(46, 77, 239)"],   # Blue (start)
    [0.5, "rgb(237, 254, 233)"], # Light greenish-white (midpoint)
    [1.0, "rgb(216, 99, 51)"]    # Orange-red (end)
]

## the following two are needed if running in jupyter notebooks
#import ipywidgets as widgets
#from IPython.display import display

@dataclass
class V_values: # this will just hold the values for couplings...
    i_e: float
    e_i: float
    i_p: float
    p_i: float
    i_f: float
    f_i: float
    p_f: float
    f_p: float
    e_f: float
    f_e: float
    e_p: float
    p_e: float

    def construct_V_matrix(self):
        V_matrix = np.array([
            [0, self.i_e, self.i_p, self.i_f],
            [self.e_i, 0, self.e_p, self.e_f],
            [self.p_i, self.p_e, 0, self.p_f],
            [self.f_i, self.f_e, self.f_p, 0]
        ])
        return V_matrix # this is the matrix actually used in the calculation of coupled surfaces

@dataclass
class Coordinate: # used for both p and q coordinates 
    angle: float # the angle from the center line, connecting the initial (i) and final (f) states
    unit: np.ndarray # the unit vector... this determines the direction in which p or q runs
    color: str # when plotting, this is used to determine the color to render the vector as.
    #vector: np.ndarray  # the x and y coordinates associated with the vector
    #length = float # the length of the vector
    #k: float # 

    def get_unit_vector(self, mode="degrees", coord="q"):
        if mode == "degrees":
            angle = np.radians(self.angle)
        self.unit = np.array([np.cos(angle), np.sin(angle)])
        if coord == "p":
            self.unit[1] *= -1
    
    # find the vector, based on the length provided
    #def get_vector(self, mode="degrees", coord="p"):
    #    if mode == "degrees":
    #        angle = np.aradians(self.angle)
    #    self.unit = np.array([np.cos(angle), np.sin(angle)])
    #    self.vector = np.array([np.cos(angle), np.sin(angle)]) * self.length
    #    if coord == "p":
    #        self.unit[1] *= -1
    #        self.vector[1] *= -1

@dataclass
class Spring: # this is the spring constants along p and q it will be a property of the uncoupled PES in units of eV/\AA^2
    p: float # for the proton transfer coordinate... effective spring constant
    q: float # for the electron transfer coordinate, effective spring constant

@dataclass
class UncoupledPES: # information for each of the uncoupled surfaces: initial (i), electron transfer (e), proton transfer (a), and final (f)
    location: np.ndarray
    energy: float = 0
    k: Spring = field(default_factory=lambda: Spring(p=1, q=1))
    z: np.ndarray = field(default_factory=lambda: np.zeros((100, 100)))

@dataclass
class CoupledPES: # the final results of electronic coupling. 
    z: np.ndarray = field(default_factory=lambda: np.zeros((100, 100)))
    TS: np.ndarray = field(default_factory=lambda: np.zeros((2, 2)))
    EA: float = 0.0
    path: np.ndarray = field(default_factory=lambda: np.zeros((2, 100)))

    def find_TS(self):
        # Placeholder for TS finding method
        pass

    def find_rxn_path(self):
        # Placeholder for reaction path finding method
        pass

#
# This is the main deal
#
@dataclass
class PCET_model:
    d: float = 1.0 # this will be the default value, but can allow the user for a distance in \AA
    q: Coordinate = field(default_factory=lambda: Coordinate(angle=45, unit=np.array([0.7071067812, 0.7071067812]), color="skyblue"))
    p: Coordinate = field(default_factory=lambda: Coordinate(angle=45, unit=np.array([0.7071067812, -0.7071067812]), color="darkorange"))
    
    # for proper plotting, need to calculate the x and y arrays correctly, based off the q and p coordinate locations...
    x: np.ndarray = field(init=False)  # Will be initialized in __post_init__
    y: np.ndarray = field(init=False)   

    i: UncoupledPES = field(default_factory=lambda: UncoupledPES(location=np.array([0,0])))
    e: UncoupledPES = field(default_factory=lambda: UncoupledPES(location=np.array([0.7071067812, 0.7071067812])))
    a: UncoupledPES = field(default_factory=lambda: UncoupledPES(location=np.array([0.7071067812, -0.7071067812]))) # a stands for atom
    f: UncoupledPES = field(default_factory=lambda: UncoupledPES(location=np.array([1,0])))
    
    couplings: np.ndarray = field(default_factory=lambda: np.array([
                            [0, 1, 1, 1],
                            [1, 0, 1, 1],
                            [1, 1, 0, 1],
                            [1, 1, 1, 0],
                        ]))
    
    cs0: CoupledPES = field(default_factory=CoupledPES)
    cs1: CoupledPES = field(default_factory=CoupledPES)
    cs2: CoupledPES = field(default_factory=CoupledPES)
    cs3: CoupledPES = field(default_factory=CoupledPES)
    surfaceplot: go.Figure = field(init=False)#field(default_factory=go.Figure)
    contourplot: go.Figure = field(init=False)#field(default_factory=go.Figure)

    def __post_init__(self): #these are things that happen after the model class has been instantiated.
        self.generate_xy()
        self.generate_contour_plot()
        self.generate_surface_plot()

    def generate_xy(self, xres=100, yres=100): # this generates the grid for z values to be used for both coupled and uncoupled surfaces
        # calculate the width
        h_sep = self.d # separation between minima along x
        v_sep = abs(self.q.unit[1])*self.d + abs(self.p.unit[1])*self.d # separation of minima perpendicular to x
        size_scaling = max([h_sep, v_sep]) # this is used to scale x and y ranges.

        x = np.linspace((0.5*self.d-1.01*size_scaling), (0.5*self.d+1.01*size_scaling), xres)
        y = np.linspace(-1.01*size_scaling, 1.01*size_scaling, yres)
        self.x, self.y = np.meshgrid(x, y)
        
    def get_PES_locations(self):
        self.i.location = [0, 0]
        self.f.location = [self.d, 0]

        # to find the location of the electron transfer (e) and proton transfer (a) uncoupled surfaces, we need to solve the following system of equations
        # returns scaling factors for the unit vectors
        c2 = self.d / (self.p.unit[0] - self.p.unit[1] * self.q.unit[0] / self.q.unit[1]) # for the pt location
        c1 = -1*c2 * self.p.unit[1] / self.q.unit[1] # for the et location
        
        #scale the unit vectors by the appropriate amount 
        self.e.location = [c1 * self.q.unit[0], c1 * self.q.unit[1]]
        self.a.location = [c2 * self.p.unit[0], c2 * self.p.unit[1]]

    def generate_uncoupled_surface_z(self): # gets the z-values for the uncoupled surfaces
        
        norm_q = np.linalg.norm(self.e.location) # since the location of the uncoupled e surface is the magnitude we want...
        
        for surface in [self.i, self.e, self.a, self.f]:
            # first, we construct the q and p coordinates, since the spring constants are defined along these coordinates
            q_coord = self.q.unit[0] * (self.x - surface.location[0]) + self.q.unit[1] * (self.y - surface.location[1])
            p_coord = self.p.unit[0] * (self.x - surface.location[0]) + self.p.unit[1] * (self.y - surface.location[1])

            # then we calculate the energy (z values) for the coordinates
            #surface.z = surface.k.q / norm_q**2 * q_coord**2 + surface.k.p * p_coord**2 + surface.energy  # for normalized along q
            surface.z = surface.k.q * q_coord**2 + surface.k.p * p_coord**2 + surface.energy # not normalized along q
        
    def generate_coupled_surface_zs(self):
        uncoupled_surfaces = [self.i.z, self.e.z, self.a.z, self.f.z]
        n_surfaces = len(uncoupled_surfaces)
        
        H = np.zeros((n_surfaces, n_surfaces, self.x.shape[0], self.x.shape[1]))
        for i in range(n_surfaces):
            H[i, i] = uncoupled_surfaces[i]
        for i in range(n_surfaces):
            for j in range(i + 1, n_surfaces):
                H[i, j] = H[j, i] = self.couplings[i, j]
        
        coupled_surfaces = np.zeros((n_surfaces, self.x.shape[0], self.x.shape[1]))
        for i in range(self.x.shape[0]):
            for j in range(self.x.shape[1]):
                H_ij = H[:, :, i, j]
                eigenvalues, _ = np.linalg.eigh(H_ij)
                coupled_surfaces[:, i, j] = eigenvalues
    
        self.cs0.z = coupled_surfaces[0]
        self.cs1.z = coupled_surfaces[1]
        self.cs2.z = coupled_surfaces[2]
        self.cs3.z = coupled_surfaces[3]

    def generate_contour_plot(self, zs=None):
        if zs is None:
            zs = [self.cs0.z] # the default is to plot just the ground state
        
        # sort these so that the highest energy will be first...

        sorted_surfaces = sorted(zs, key=lambda x: np.min(x), reverse = True) # sort arrays by minimum z-value, but sort in reverse order

        contour_plot = make_subplots(rows = len(zs), cols = 1)
        for i, z in enumerate(sorted_surfaces):

            z_min = np.min(z)
            z_max = np.max(z)

            # calculate the width
            h_sep = self.d
            v_sep = abs(self.q.unit[1])*self.d + abs(self.p.unit[1])*self.d
            size_scaling = max([h_sep, v_sep]) # this is used to scale x and y ranges.
                        
            contour_plot.add_trace(go.Contour(
                    colorscale = light_temperature_map, 
                    x=self.x[0], 
                    y=self.y[:,0],
                    z=z,
                    showscale = False,
                    contours=dict(
                        start=z_min,  # Start value for contours
                        end=z_max,      # End value for contours
                        size=(z_max - z_min)/25     # Spacing between contours
                    ),
                    zmin=z_min,  # Set the initial min value for the color scale
                    zmax=z_max,   # Set the initial max value for the color scale
                ), row = i+1, col = 1,
            )
            
            # now, let us add the p and q vectors...
            #first q
            contour_plot.add_scatter(x = [self.i.location[0], self.e.location[0], None, self.a.location[0], self.f.location[0]], 
                                    y = [self.i.location[1], self.e.location[1], None, self.a.location[1], self.f.location[1]],
                                   mode = "markers+lines", 
                                   marker = dict(color = "white"),
                                    line = dict(color = self.q.color),
                                     showlegend = False,
                                     row = i+1, col = 1,
                                   )
            #then p
            contour_plot.add_scatter(x = [self.i.location[0], self.a.location[0], None, self.e.location[0], self.f.location[0]], 
                                    y = [self.i.location[1], self.a.location[1], None, self.e.location[1], self.f.location[1]],
                                   mode = "markers+lines", 
                                   marker = dict(color = "white"),
                                    line = dict(color = self.p.color),
                                     showlegend = False,
                                     row = i+1, col = 1,
                                   )

            contour_plot.add_scatter(x = [self.i.location[0],self.f.location[0],],
                                     y = [self.i.location[1],self.f.location[1]],
                                     mode = "lines+markers",
                                     marker = dict(color = "white"),
                                    line = dict(color = "white"),
                                     showlegend = False,
                                     row = i+1, col = 1,
                                    )
                                 
        contour_plot.update_xaxes(title = "concerted transfer coordinate / \u00C5")
        contour_plot.update_yaxes(title = "charge-separation coordinate / \u00C5")
        contour_plot.update_layout(
            margin = dict(l=10, t=10, r=10, b=10),
            width = 400, height = 400*len(zs), # make this as tall as needed
            template = "simple_white",
        )
        self.contourplot = contour_plot
        return contour_plot
    
    def generate_surface_plot(self, zs=None):
        '''
        The optional keyword zs can accept a list of surfaces to plot
        '''
        # calculate the width
        h_sep = self.d
        v_sep = abs(self.q.unit[1])*self.d + abs(self.p.unit[1])*self.d
        size_scaling = max([h_sep, v_sep]) # this is used to scale x and y ranges.
        
        
        if zs is None:
            zs = [self.cs0.z]
        #print(zs)
        
        sorted_surfaces = sorted(zs, key=lambda x: np.min(x), reverse = True) # sort arrays by minimum z-value, but sort in reverse order

        # build specs lists...
        specs_list = []
        for i in range(len(zs)):
            specs_list.append([{"type":"surface"}])

        surface_plot = make_subplots(rows = len(zs), cols = 1, specs = specs_list)
        

        for i, z in enumerate(sorted_surfaces):
            surface_plot.add_trace(go.Surface(
                colorscale = light_temperature_map,
                x=self.x, 
                y=self.y,
                z=z,
                showscale = False,
            ), row = i+1, col = 1
            )
        surface_plot.update_layout(
            margin = dict(l=10, t=10, r=10, b=10),
            width = 400, height = 400*len(zs),
            template = "simple_white",
            scene=dict(
                aspectmode="cube", # keepx the length of the x,y,and z axes on the page the same length
                # to determine the size of the plot... need to figure out the full range of separation between wells along x and y, and scale by that
                xaxis=dict(title = "concerted transfer coordinate / \u00C5"),
                yaxis=dict(title = "charge-separation coordinate / \u00C5"),
                zaxis=dict(title = "energy / eV", range=[np.min(zs), np.max(zs)]),
            )
        )
        self.surfaceplot = surface_plot
        return surface_plot
